<?php
namespace Mewz\WCAS\Compatibility\Actions;

use Mewz\Framework\Base\Action;
use Mewz\WCAS\Actions\Front\ProductLimits;
use Mewz\WCAS\Models\AttributeStock;
use Mewz\WCAS\Util;

class WPLister extends Action
{
	public function __hooks()
	{
		// override product stock everywhere (frontend and backend)
		add_filter('mewz_wcas_limit_product_stock', '__return_true');
		add_action('wplister_revise_inventory_status', [$this, 'enable_limit_product_stock'], 0);
		add_action('wpla_product_has_changed', [$this, 'enable_limit_product_stock'], 0);

		// account for attribute stock when calculating total variable product stock
		add_filter('get_post_metadata', [$this, 'filter_product_stock_meta'], 0, 4);

		// trigger product quantity updated hooks when attribute stock is saved/changed/deleted
		add_action('mewz_attribute_stock_saved', [$this, 'mewz_attribute_stock_saved'], 10, 3);
		add_action('save_post_mewz_attribute_stock', [$this, 'trigger_product_updates'], 20);
		add_action('mewz_wcas_match_sets_before_delete', [$this, 'trigger_product_updates']);

		// trigger product quantity updated hooks when attributes are deleted
		add_action('delete_term', [$this, 'delete_term'], 0, 3);
		add_action('woocommerce_attribute_deleted', [$this, 'deleted_attribute'], 0);

		// sync attribute stock when WP-Lister imports an order
		add_action('wplister_after_create_order', [$this, 'after_create_order']);
		add_action('wple_after_create_order', [$this, 'after_create_order']);
		add_action('wpla_after_create_order', [$this, 'after_create_order']);
	}

	public function enable_limit_product_stock()
	{
		if (!class_exists(ProductLimits::class, false)) {
			$this->plugin->loader->action(ProductLimits::class);

			add_action('wplister_revise_inventory_status', [$this, 'disable_limit_product_stock'], 9999);
			add_action('wpla_product_has_changed', [$this, 'disable_limit_product_stock'], 9999);
	    }
	}

	public function disable_limit_product_stock()
	{
		ProductLimits::$enabled = false;
	}

	public function filter_product_stock_meta($value, $object_id, $meta_key, $single)
	{
		if ($meta_key === '_stock' && $single) {
			remove_filter('get_post_metadata', [$this, 'filter_product_stock_meta'], 0);

			$product = wc_get_product($object_id);

			if ($product && $product instanceof \WC_Product_Variation) {
				$value = $product->get_stock_quantity();
			}

			add_filter('get_post_metadata', [$this, 'filter_product_stock_meta'], 0, 4);
		}

	    return $value;
	}

	/**
	 * @param AttributeStock $stock
	 * @param string $operation
	 * @param array $changes
	 */
	public function mewz_attribute_stock_saved($stock, $operation, $changes)
	{
		$this->trigger_product_updates($stock->id());
	}

	/**
	 * @param int $term_id
	 * @param int $tt_id
	 * @param string $taxonomy
	 */
	public function delete_term($term_id, $tt_id, $taxonomy)
	{
		if (strpos($taxonomy, 'pa_') !== 0) return;

		$stock_ids = Util\Matches::query()
			->where('r.attribute_id', Util\Attributes::get_attribute_id($taxonomy))
			->where('r.term_id', $term_id)
			->col('s.stock_id');

		foreach ($stock_ids as $stock_id) {
			$this->trigger_product_updates($stock_id);
		}
	}

	/**
	 * @param int $attribute_id
	 */
	public function deleted_attribute($attribute_id)
	{
		$stock_ids = Util\Matches::query()
			->where('r.attribute_id', $attribute_id)
			->col('s.stock_id');

		foreach ($stock_ids as $stock_id) {
		    $this->trigger_product_updates($stock_id);
		}
	}

	public function after_create_order($order_id)
	{
		Util\Orders::update_order_attribute_stock($order_id, 'sync');
	}

	public function trigger_product_updates($stock_id)
	{
		$match_sets = Util\Matches::get_sets($stock_id);
		if (!$match_sets) return;

		$product_ids = [];

		// find all matching product IDs
		foreach ($match_sets as $match_set) {
			if ($match_set['multiplier'] == 0) {
				continue;
			}

			$tax_queries = [];

			// build tax query
			foreach ($match_set['rows'] as $row) {
				$taxonomy = Util\Attributes::get_attribute_name($row['attribute'], true);
				if (!$taxonomy) continue;

				$tax_query['taxonomy'] = $taxonomy;

				if ($row['term']) {
					$tax_query['terms'] = $row['term'];
				}

				$tax_queries[] = $tax_query;
			}

			// find all matching product IDs
			$found_ids = get_posts([
				'post_type' => 'product',
				'post_status' => ['publish', 'private'],
				'fields' => 'ids',
				'tax_query' => $tax_queries,
				'numberposts' => -1,
				'update_post_meta_cache' => false,
				'update_post_term_cache' => false,
				'no_found_rows' => true,
			]);

			if ($found_ids) {
				foreach ($found_ids as $found_id) {
					$product_ids[$found_id] = true;
				}
			}
		}

		if (!$product_ids) return;

		foreach ($product_ids as $product_id => $_) {
			if (!Util\Products::is_product_excluded($product_id)) {
				do_action('wplister_revise_inventory_status', $product_id);
				do_action('wpla_product_has_changed', $product_id);
			}
		}
	}
}
